using MediatR;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Persistence;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using Microsoft.Extensions.Localization;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Services;
using cuqmbr.TravelGuide.Domain.Enums;
using FluentValidation.Results;

namespace cuqmbr.TravelGuide.Application.TicketGroups.Commands.AddTicketGroup;

// TODO: Add descriptions and proper property names to validation errors

public class AddTicketGroupCommandHandler :
    IRequestHandler<AddTicketGroupCommand, TicketGroupDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly IStringLocalizer _localizer;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    public AddTicketGroupCommandHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        IStringLocalizer localizer,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
    }

    public async Task<TicketGroupDto> Handle(
        AddTicketGroupCommand request,
        CancellationToken cancellationToken)
    {
        // Check whether provided vehicle enrollments are present in datastore.
        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            if (vehicleEnrollmentGuids.Count() > vehicleEnrollments.Count)
            {
                throw new NotFoundException();
            }
        }


        // Check whether provided arrival and departure address guids
        // are used in provided vehicle enrollment and
        // and are in the correct order.
        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e =>
                        routeAddressGuids.Contains(e.Guid),
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            foreach (var t in request.Tickets)
            {
                var departureRouteAddress = routeAddresses.First(
                    ra => ra.Guid == t.DepartureRouteAddressGuid);
                var arrivalRouteAddress = routeAddresses.First(
                    ra => ra.Guid == t.ArrivalRouteAddressGuid);

                var ve = vehicleEnrollments.First(
                    e => e.Guid == t.VehicleEnrollmentGuid);

                if (departureRouteAddress.RouteId != ve.RouteId ||
                    arrivalRouteAddress.RouteId != ve.RouteId)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }

                if (departureRouteAddress.Order > arrivalRouteAddress.Order)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Check availability of free places.
        {
            // Get all tickets for vehicle enrollments requested in ticket group.
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);
            
            var ticketGroupTickets = (await _unitOfWork.TicketRepository
                .GetPageAsync(
                    e =>
                        vehicleEnrollmentGuids.Contains(e.VehicleEnrollment.Guid) &&
                        e.TicketGroup.Returned == false,
                    1, int.MaxValue, cancellationToken))
                .Items;


            // Get all vehicle enrollments requested in ticket group
            // together with vehicles.
            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    e => e.Vehicle,
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            // Get all route addresses of vehicle enrollments
            // requested in ticket group.
            var routeIds = vehicleEnrollments.Select(e => e.RouteId);

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e => routeIds.Contains(e.RouteId),
                    1, int.MaxValue, cancellationToken))
                .Items;

            // For each ticket in request.
            foreach (var requestTicket in request.Tickets)
            {
                // Get vehicle enrollment of requested ticket.
                var requestVehicleEnrollment = vehicleEnrollments.First(e =>
                    e.Guid == requestTicket.VehicleEnrollmentGuid);

                // Get bought tickets of vehicle enrollment of requested ticket.
                var tickets = ticketGroupTickets.Where(t =>
                    t.VehicleEnrollmentId == requestVehicleEnrollment.Id);

                // Get route addresses of vehicle enrollment.
                var ticketRouteAddresses = routeAddresses
                    .Where(e => e.RouteId == requestVehicleEnrollment.RouteId)
                    .OrderBy(e => e.Order);


                // Count available capacity.

                // Get total capacity in requested vehicle.
                int totalCapacity;
                var vehicle = vehicleEnrollments.First(e =>
                        e.Guid == requestTicket.VehicleEnrollmentGuid)
                    .Vehicle;
                if (vehicle.VehicleType.Equals(VehicleType.Bus))
                {
                    totalCapacity = ((Bus)vehicle).Capacity;
                }
                else if (vehicle.VehicleType.Equals(VehicleType.Aircraft))
                {
                    totalCapacity = ((Aircraft)vehicle).Capacity;
                }
                else if (vehicle.VehicleType.Equals(VehicleType.Train))
                {
                    totalCapacity = ((Train)vehicle).Capacity;
                }
                else
                {
                    throw new NotImplementedException();
                }

                int takenCapacity = 0;

                // For each bought ticket.
                foreach (var ticket in tickets)
                {
                    // Get departure and arrival route address
                    // of requested ticket.
                    var requestDepartureRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Guid == requestTicket.DepartureRouteAddressGuid);
                    var requestArrivalRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Guid == requestTicket.ArrivalRouteAddressGuid);
                    
                    // Get departure and arrival route address
                    // of bought ticket.
                    var departureRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Id == ticket.DepartureRouteAddressId);
                    var arrivalRouteAddress = ticketRouteAddresses
                        .Single(e =>
                            e.Id == ticket.ArrivalRouteAddressId);


                    // Count taken capacity in requested vehicle
                    // accounting for requested ticket
                    // departure and arrival route addresses.
                    // The algorithm is the same as vehicle enrollment
                    // time overlap check.
                    if ((requestDepartureRouteAddress.Order >= 
                            departureRouteAddress.Order &&
                        requestDepartureRouteAddress.Order <
                            arrivalRouteAddress.Order) ||
                        (requestArrivalRouteAddress.Order <=
                            arrivalRouteAddress.Order &&
                        requestArrivalRouteAddress.Order >
                            departureRouteAddress.Order) ||
                        (requestDepartureRouteAddress.Order <=
                            departureRouteAddress.Order &&
                        requestArrivalRouteAddress.Order >=
                            arrivalRouteAddress.Order))
                    {
                        takenCapacity++;
                    }
                }

                var availableCapacity = totalCapacity - takenCapacity;

                if (availableCapacity <= 0)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Calculate travel time and cost.

        var ticketsDetails = new List<(short order, DateTimeOffset departureTime,
            DateTimeOffset arrivalTime, decimal cost, Currency currency)>();

        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e =>
                        routeAddressGuids.Contains(e.Guid),
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            var vehicleEnrollmentIds = vehicleEnrollments.Select(ve => ve.Id);

            var allRouteAddressDetails = (await _unitOfWork
                .RouteAddressDetailRepository.GetPageAsync(
                    e => vehicleEnrollmentIds.Contains(e.VehicleEnrollmentId),
                    e => e.RouteAddress,
                    1, int.MaxValue, cancellationToken))
                .Items;


            foreach (var t in request.Tickets.OrderBy(t => t.Order))
            {
                var ve = vehicleEnrollments.First(
                    e => e.Guid == t.VehicleEnrollmentGuid);

                var departureRouteAddressId = routeAddresses.First(
                        ra => ra.Guid == t.DepartureRouteAddressGuid)
                    .Id;
                var arrivalRouteAddressId = routeAddresses.First(
                        ra => ra.Guid == t.ArrivalRouteAddressGuid)
                    .Id;

                var verad = allRouteAddressDetails
                    .Where(arad => arad.VehicleEnrollmentId == ve.Id)
                    .OrderBy(rad => rad.RouteAddress.Order)
                    .TakeWhile(rad => rad.Id != arrivalRouteAddressId);


                // TODO: This counts departure address stop time which is
                // not wrong but may be not desired.
                var timeToDeparture = verad
                    .TakeWhile(rad => rad.Id != departureRouteAddressId)
                    .Aggregate(TimeSpan.Zero, (sum, next) =>
                        sum + next.TimeToNextAddress + next.CurrentAddressStopTime);

                var departureTime = ve.DepartureTime.Add(timeToDeparture);


                var timeToArrival = verad.Aggregate(TimeSpan.Zero, (sum, next) =>
                    sum + next.TimeToNextAddress + next.CurrentAddressStopTime);

                var arrivalTime = ve.DepartureTime.Add(timeToArrival);


                var costToDeparture = verad
                    .TakeWhile(rad => rad.Id != departureRouteAddressId)
                    .Aggregate((decimal)0, (sum, next) =>
                        sum + next.CostToNextAddress);

                var costToArrival = verad
                    .Aggregate((decimal)0, (sum, next) =>
                        sum + next.CostToNextAddress);

                var cost = costToArrival - costToDeparture;


                ticketsDetails.Add(
                    (t.Order, departureTime, arrivalTime, cost, ve.Currency));
            }
        }

        // Check whether there are overlaps in ticket departure/arrival times.
        {
            for (int i = 1; i < ticketsDetails.Count; i++)
            {
                var previousTd = ticketsDetails[i - 1];
                var currentTd = ticketsDetails[i];

                if (previousTd.arrivalTime >= currentTd.departureTime)
                {
                    throw new ValidationException(
                        new List<ValidationFailure>
                        {
                            new()
                            {
                                PropertyName = nameof(request.Tickets)
                            }
                        });
                }
            }
        }

        // Create entity and insert into a datastore.

        var ticketsCostDetails = new List<(short order,
            decimal cost, Currency currency)>();

        TimeSpan travelTime;

        {
            travelTime =
                ticketsDetails.OrderBy(td => td.order).Last().arrivalTime -
                ticketsDetails.OrderBy(td => td.order).First().departureTime;

            foreach (var td in ticketsDetails)
            {
                var initialCurrency = td.currency;
                var convertedCurrency =
                    _sessionCurrencyService.Currency != Currency.Default ?
                    _sessionCurrencyService.Currency :
                    initialCurrency;

                var cost = td.cost;

                var convertedCost = await _currencyConverterService
                    .ConvertAsync(cost, initialCurrency,
                        convertedCurrency, cancellationToken);

                ticketsCostDetails.Add((td.order, convertedCost, convertedCurrency));
            }
        }

        {
            var vehicleEnrollmentGuids =
                request.Tickets.Select(t => t.VehicleEnrollmentGuid);

            var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
                .GetPageAsync(
                    e => vehicleEnrollmentGuids.Contains(e.Guid),
                    1, vehicleEnrollmentGuids.Count(), cancellationToken))
                .Items;


            var routeAddressGuids =
                request.Tickets.Select(t => t.DepartureRouteAddressGuid).Concat(
                    request.Tickets.Select(t => t.ArrivalRouteAddressGuid));

            var routeAddresses = (await _unitOfWork.RouteAddressRepository
                .GetPageAsync(
                    e => routeAddressGuids.Contains(e.Guid),
                    e => e.Address.City.Region.Country,
                    1, routeAddressGuids.Count(), cancellationToken))
                .Items;


            var entity = new TicketGroup()
            {
                PassangerFirstName = request.PassangerFirstName,
                PassangerLastName = request.PassangerLastName,
                PassangerPatronymic = request.PassangerPatronymic,
                PassangerSex = request.PassangerSex,
                PassangerBirthDate = request.PassangerBirthDate,
                PurchaseTime = request.PurchaseTime,
                Returned = request.Returned,
                TravelTime = travelTime,
                Tickets = request.Tickets.Select(
                    t =>
                    {
                        var ve = vehicleEnrollments.First(
                            ve => ve.Guid == t.VehicleEnrollmentGuid);


                        var departureRouteAddress = routeAddresses.First(
                                ra => ra.Guid == t.DepartureRouteAddressGuid);
                        var arrivalRouteAddress = routeAddresses.First(
                                ra => ra.Guid == t.ArrivalRouteAddressGuid);


                        var costDetail = ticketsCostDetails
                            .SingleOrDefault(td => td.order == t.Order);


                        return new Ticket()
                        {
                            DepartureRouteAddressId = departureRouteAddress.Id,
                            DepartureRouteAddress = departureRouteAddress,
                            ArrivalRouteAddressId = arrivalRouteAddress.Id,
                            ArrivalRouteAddress = arrivalRouteAddress,
                            Order = t.Order,
                            Cost = costDetail.cost,
                            Currency = costDetail.currency,
                            VehicleEnrollmentId = ve.Id
                        };
                    })
                    .ToArray()
            };

            entity = await _unitOfWork.TicketGroupRepository.AddOneAsync(
                entity, cancellationToken);

            await _unitOfWork.SaveAsync(cancellationToken);
            _unitOfWork.Dispose();

            return _mapper.Map<TicketGroupDto>(entity);
        }
    }
}
