using MediatR;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Persistence;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using cuqmbr.TravelGuide.Domain.Entities;
using AutoMapper;
using QuikGraph;
using QuikGraph.Algorithms;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Services;

namespace cuqmbr.TravelGuide.Application
    .VehicleEnrollmentSearch.Queries.SearchShortest;

// TODO: Refactor.
// TODO: Add configurable time between transfers.
public class SearchShortestQueryHandler :
    IRequestHandler<SearchShortestQuery, VehicleEnrollmentSearchDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    public SearchShortestQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
    }

    public async Task<VehicleEnrollmentSearchDto> Handle(
        SearchShortestQuery request,
        CancellationToken cancellationToken)
    {
        // Get related data

        var zeroTime = TimeOnly.FromTimeSpan(TimeSpan.Zero);
        var departureDate =
            new DateTimeOffset(request.DepartureDate, zeroTime, TimeSpan.Zero);

        var range = TimeSpan.FromDays(3);
        var vehicleEnrollments = (await _unitOfWork.VehicleEnrollmentRepository
            .GetPageAsync(
                e => 
                    e.DepartureTime >= departureDate.Subtract(range) &&
                    e.DepartureTime <= departureDate.Add(range) &&
                    request.VehicleTypes.Contains(e.Vehicle.VehicleType),
                e => e.Route,
                1, int.MaxValue, cancellationToken))
            .Items;

        if (vehicleEnrollments.Count == 0)
        {
            throw new NotFoundException();
        }

        var vehicleEnrollmentIds = vehicleEnrollments.Select(e => e.Id);
        var routeAddressDetails = (await _unitOfWork.RouteAddressDetailRepository
            .GetPageAsync(
                e => vehicleEnrollmentIds.Contains(e.VehicleEnrollmentId),
                e => e.RouteAddress.Address.City.Region.Country,
                1, int.MaxValue, cancellationToken))
            .Items;


        // Hydrate vehicle enrollments with route address details

        foreach (var vehicleEnrollment in vehicleEnrollments)
        {
            vehicleEnrollment.RouteAddressDetails = routeAddressDetails
                .Where(e => e.VehicleEnrollmentId == vehicleEnrollment.Id)
                .OrderBy(e => e.RouteAddress.Order)
                .ToArray();
        }


        // Creat and fill graph data structure

        var graph = new AdjacencyGraph<
            Address, TaggedEdge<Address, RouteAddressDetail>>();

        foreach (var vehicleEnrollment in vehicleEnrollments)
        {
            var vehicleEnrollmentRouteAddressDetails = routeAddressDetails
                .Where(e => e.VehicleEnrollmentId == vehicleEnrollment.Id)
                .OrderBy(e => e.RouteAddress.Order);

            for (int i = 1; i < vehicleEnrollmentRouteAddressDetails.Count(); i++)
            {
                var sourceRouteAddressDetail =
                    vehicleEnrollmentRouteAddressDetails.ElementAt(i-1);
                var targetRouteAddressDetail =
                    vehicleEnrollmentRouteAddressDetails.ElementAt(i);

                var sourceAddress =
                    sourceRouteAddressDetail.RouteAddress.Address;
                var targetAddress =
                    targetRouteAddressDetail.RouteAddress.Address;

                var weight = sourceRouteAddressDetail.CostToNextAddress;
                
                graph.AddVerticesAndEdge(
                    new TaggedEdge<Address, RouteAddressDetail>(
                        sourceAddress, targetAddress, sourceRouteAddressDetail));
            }
        }


        // Find paths

        var departureAddress = routeAddressDetails
            .First(e => e.RouteAddress.Address.Guid == request.DepartureAddressGuid)
            .RouteAddress.Address;
        var arrivalAddress = routeAddressDetails
            .First(e => e.RouteAddress.Address.Guid == request.ArrivalAddressGuid)
            .RouteAddress.Address;


        Func<TaggedEdge<Address, RouteAddressDetail>, double> weightByCost =
            edge => (double)edge.Tag.CostToNextAddress;

        Func<TaggedEdge<Address, RouteAddressDetail>, double> weightByTime =
            edge =>
                edge.Tag.TimeToNextAddress.Ticks +
                edge.Tag.CurrentAddressStopTime.Ticks;

        Func<TaggedEdge<Address, RouteAddressDetail>, double> edgeWeight =
            _ => 0;

        if (request.ShortestByCost && request.ShortestByTime)
        {
            edgeWeight = edge => weightByCost(edge) + weightByTime(edge);
        }
        else if (request.ShortestByCost)
        {
            edgeWeight = edge => weightByCost(edge);
        }
        else if (request.ShortestByTime)
        {
            edgeWeight = edge => weightByTime(edge);
        }


        var tryGetPaths = graph.ShortestPathsDijkstra(edgeWeight, departureAddress);


        // Create and hydrate a DTO object

        var vehicleEnrollmentDtos =
            new List<VehicleEnrollmentSearchVehicleEnrollmentDto>();

        var totalTravelTime = TimeSpan.Zero;
        var totalCost = (decimal)0;

        if (tryGetPaths(arrivalAddress, out var path))
        {
            var firstRouteAddressId = path.First().Tag.RouteAddressId;
            long lastRouteAddressId;
            Guid lastRouteAddressGuid;

            var addressDtos = new List<VehicleEnrollmentSearchAddressDto>();
            var addressOrder = (short)1;
            var enrollmentTravelTime = TimeSpan.Zero;
            var enrollmentCost = (decimal)0;
            var enrollmentOrder = (short)1;

            Address source;
            Address target;
            RouteAddressDetail tag;
            RouteAddressDetail nextTag;

            for (int i = 0; i < path.Count() - 1; i++)
            {
                source = path.Select(e => e.Source).ElementAt(i);
                tag = path.Select(e => e.Tag).ElementAt(i);
                nextTag = path.Select(e => e.Tag).ElementAt(i+1);


                totalTravelTime +=
                    tag.TimeToNextAddress + tag.CurrentAddressStopTime;
                enrollmentTravelTime +=
                    tag.TimeToNextAddress + tag.CurrentAddressStopTime;

                totalCost += tag.CostToNextAddress;
                enrollmentCost += tag.CostToNextAddress;

                addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
                {
                    Uuid = source.Guid,
                    Name = source.Name,
                    Longitude = source.Latitude,
                    Latitude = source.Longitude,
                    CountryUuid = source.City.Region.Country.Guid,
                    CountryName = source.City.Region.Country.Name,
                    RegionUuid = source.City.Region.Guid,
                    RegionName = source.City.Region.Name,
                    CityUuid = source.City.Guid,
                    CityName = source.City.Name,
                    TimeToNextAddress = tag.TimeToNextAddress,
                    CostToNextAddress = tag.CostToNextAddress,
                    CurrentAddressStopTime = tag.CurrentAddressStopTime,
                    Order = addressOrder,
                    RouteAddressUuid = tag.RouteAddress.Guid
                });

                addressOrder++;


                // First address after transfer
                if (nextTag.VehicleEnrollmentId != tag.VehicleEnrollmentId)
                {
                    target = path.Select(e => e.Target).ElementAt(i);

                    lastRouteAddressGuid = vehicleEnrollments
                        .Single(e => e.Id == tag.VehicleEnrollmentId)
                        .RouteAddressDetails
                        .Select(e => e.RouteAddress)
                        .OrderBy(e => e.Order)
                        .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                        .Take(2)
                        .ElementAt(1)
                        .Guid;

                    addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
                    {
                        Uuid = target.Guid,
                        Name = target.Name,
                        Longitude = target.Latitude,
                        Latitude = target.Longitude,
                        CountryUuid = target.City.Region.Country.Guid,
                        CountryName = target.City.Region.Country.Name,
                        RegionUuid = target.City.Region.Guid,
                        RegionName = target.City.Region.Name,
                        CityUuid = target.City.Guid,
                        CityName = target.City.Name,
                        TimeToNextAddress = TimeSpan.Zero,
                        CostToNextAddress = 0,
                        CurrentAddressStopTime = TimeSpan.Zero,
                        Order = addressOrder,
                        RouteAddressUuid = lastRouteAddressGuid
                    });

                    lastRouteAddressId = vehicleEnrollments
                        .Single(e => e.Id == tag.VehicleEnrollmentId)
                        .RouteAddressDetails
                        .Select(e => e.RouteAddress)
                        .OrderBy(e => e.Order)
                        .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                        .Take(2)
                        .ElementAt(1)
                        .Id;

                    vehicleEnrollmentDtos.Add(
                        new VehicleEnrollmentSearchVehicleEnrollmentDto()
                        {
                            DepartureTime = tag.VehicleEnrollment
                                .GetDepartureTime(firstRouteAddressId),
                            ArrivalTime = tag.VehicleEnrollment
                                .GetArrivalTime(lastRouteAddressId),
                            TravelTime = tag.VehicleEnrollment
                                .GetTravelTime(firstRouteAddressId,
                                    lastRouteAddressId),
                            TimeMoving = tag.VehicleEnrollment
                                .GetTimeMoving(firstRouteAddressId,
                                    lastRouteAddressId),
                            TimeInStops = tag.VehicleEnrollment
                                .GetTimeInStops(firstRouteAddressId,
                                    lastRouteAddressId),
                            NumberOfStops = tag.VehicleEnrollment
                                .GetNumberOfStops(firstRouteAddressId,
                                    lastRouteAddressId),
                            Currency = tag.VehicleEnrollment.Currency.Name,
                            Cost = tag.VehicleEnrollment
                                .GetCost(firstRouteAddressId,
                                    lastRouteAddressId),
                            VehicleType = source.VehicleType.Name,
                            Uuid = tag.VehicleEnrollment.Guid,
                            Order = enrollmentOrder,
                            Addresses = addressDtos
                        });

                    firstRouteAddressId = nextTag.RouteAddressId;

                    addressDtos = new List<VehicleEnrollmentSearchAddressDto>();
                    addressOrder = (short)1;
                    enrollmentTravelTime = TimeSpan.Zero;
                    enrollmentCost = (decimal)0;
                    enrollmentOrder++;
                }
            }

            source = path.Select(e => e.Source).Last();
            target = path.Select(e => e.Target).Last();
            tag = path.Select(e => e.Tag).Last();
            nextTag = path.Select(e => e.Tag).Last();

            addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
            {
                Uuid = source.Guid,
                Name = source.Name,
                Longitude = source.Latitude,
                Latitude = source.Longitude,
                CountryUuid = source.City.Region.Country.Guid,
                CountryName = source.City.Region.Country.Name,
                RegionUuid = source.City.Region.Guid,
                RegionName = source.City.Region.Name,
                CityUuid = source.City.Guid,
                CityName = source.City.Name,
                TimeToNextAddress = tag.TimeToNextAddress,
                CostToNextAddress = tag.CostToNextAddress,
                CurrentAddressStopTime = tag.CurrentAddressStopTime,
                Order = addressOrder,
                RouteAddressUuid = tag.RouteAddress.Guid
            });

            lastRouteAddressGuid = vehicleEnrollments
                .Single(e => e.Id == tag.VehicleEnrollmentId)
                .RouteAddressDetails
                .Select(e => e.RouteAddress)
                .OrderBy(e => e.Order)
                .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                .Take(2)
                .ElementAt(1)
                .Guid;

            addressOrder++;
            addressDtos.Add(new VehicleEnrollmentSearchAddressDto()
            {
                Uuid = target.Guid,
                Name = target.Name,
                Longitude = target.Latitude,
                Latitude = target.Longitude,
                CountryUuid = target.City.Region.Country.Guid,
                CountryName = target.City.Region.Country.Name,
                RegionUuid = target.City.Region.Guid,
                RegionName = target.City.Region.Name,
                CityUuid = target.City.Guid,
                CityName = target.City.Name,
                TimeToNextAddress = TimeSpan.Zero,
                CostToNextAddress = 0,
                CurrentAddressStopTime = TimeSpan.Zero,
                Order = addressOrder,
                RouteAddressUuid = lastRouteAddressGuid
            });

            lastRouteAddressId = vehicleEnrollments
                .Single(e => e.Id == tag.VehicleEnrollmentId)
                .RouteAddressDetails
                .Select(e => e.RouteAddress)
                .OrderBy(e => e.Order)
                .SkipWhile(e => e.Order != tag.RouteAddress.Order)
                .Take(2)
                .ElementAt(1)
                .Id;

            vehicleEnrollmentDtos.Add(
                new VehicleEnrollmentSearchVehicleEnrollmentDto()
                {
                    DepartureTime = tag.VehicleEnrollment
                        .GetDepartureTime(firstRouteAddressId),
                    ArrivalTime = tag.VehicleEnrollment
                        .GetArrivalTime(lastRouteAddressId),
                    TravelTime = tag.VehicleEnrollment
                        .GetTravelTime(firstRouteAddressId,
                            lastRouteAddressId),
                    TimeMoving = tag.VehicleEnrollment
                        .GetTimeMoving(firstRouteAddressId,
                            lastRouteAddressId),
                    TimeInStops = tag.VehicleEnrollment
                        .GetTimeInStops(firstRouteAddressId,
                            lastRouteAddressId),
                    NumberOfStops = tag.VehicleEnrollment
                        .GetNumberOfStops(firstRouteAddressId,
                            lastRouteAddressId),
                    Currency = tag.VehicleEnrollment.Currency.Name,
                    Cost = tag.VehicleEnrollment
                        .GetCost(firstRouteAddressId,
                            lastRouteAddressId),
                    VehicleType = source.VehicleType.Name,
                    Uuid = tag.VehicleEnrollment.Guid,
                    Order = enrollmentOrder,
                    Addresses = addressDtos
                });
        }
        else
        {
            throw new NotFoundException();
        }


        foreach (var vehicleEnrollmentDto in vehicleEnrollmentDtos)
        {
            foreach (var addressDto in vehicleEnrollmentDto.Addresses)
            {
                addressDto.CostToNextAddress = await _currencyConverterService
                    .ConvertAsync(addressDto.CostToNextAddress,
                        vehicleEnrollments
                            .First(e => e.Guid == vehicleEnrollmentDto.Uuid)
                            .Currency,
                        _sessionCurrencyService.Currency, cancellationToken);
            }

            vehicleEnrollmentDto.Currency = vehicleEnrollmentDto.Currency;
            vehicleEnrollmentDto.Cost = vehicleEnrollmentDto.Addresses
                .Aggregate((decimal)0,
                    (sum, next) => sum += next.CostToNextAddress);
        }

        var cost = vehicleEnrollmentDtos
                .Aggregate((decimal)0,
                    (sum, next) => sum += next.Cost);

        var departureTime = vehicleEnrollmentDtos
            .OrderBy(e => e.Order).First().DepartureTime;
        var arrivalTime = vehicleEnrollmentDtos
            .OrderBy(e => e.Order).Last().ArrivalTime;
        var timeInStops = vehicleEnrollmentDtos
            .Aggregate(TimeSpan.Zero, (sum, next) => sum += next.TimeInStops);
        var numberOfTransfers = vehicleEnrollmentDtos.Count() - 1;

        return new VehicleEnrollmentSearchDto()
        {
            DepartureTime = departureTime,
            ArrivalTime = arrivalTime,
            TravelTime = arrivalTime - departureTime,
            TimeInStops = timeInStops,
            NumberOfTransfers = numberOfTransfers,
            Enrollments = vehicleEnrollmentDtos
        };
    }
}
