using MediatR;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Exceptions;
using FluentValidation.Results;
using Microsoft.Extensions.Localization;
using cuqmbr.TravelGuide.Domain.Entities;

namespace cuqmbr.TravelGuide.Application.VehicleEnrollments
    .Commands.UpdateVehicleEnrollment;

public class UpdateVehicleEnrollmentCommandHandler :
    IRequestHandler<UpdateVehicleEnrollmentCommand, VehicleEnrollmentDto>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;
    private readonly IStringLocalizer _localizer;

    public UpdateVehicleEnrollmentCommandHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        IStringLocalizer localizer)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _localizer = localizer;
    }

    public async Task<VehicleEnrollmentDto> Handle(
        UpdateVehicleEnrollmentCommand request,
        CancellationToken cancellationToken)
    {
        // TODO: Check for tickets bought for this enrollment.
        // Decide whether allow or not to perform update action.

        var entity = await _unitOfWork.VehicleEnrollmentRepository.GetOneAsync(
            e => e.Guid == request.Guid, e => e.RouteAddressDetails,
            cancellationToken);

        if (entity == null)
        {
            throw new NotFoundException();
        }

        // Check that request has the same Route Addresses
        // as a route data from datastore.

        var route = await _unitOfWork.RouteRepository.GetOneAsync(
            e => e.Id == entity.RouteId, e => e.RouteAddresses,
            cancellationToken);

        var sameRouteAddresses = route.RouteAddresses.All(
            ra => request.RouteAddressDetails.Any(
                rad => rad.RouteAddressGuid == ra.Guid));

        if (!sameRouteAddresses)
        {
            throw new NotFoundException(
                $"Not all route addresses are found in a datastore.");
        }


        // Check vehicle enrollments that might overlap with updated one.
        // Exclude this vehicle enrollment.

        var requestDepartureTime = request.DepartureTime;

        var requestTravelTime =
            request.RouteAddressDetails.Aggregate(
                TimeSpan.Zero, (sum, rad) => sum +
                    rad.TimeToNextAddress + rad.CurrentAddressStopTime);

        var requestArrivalTime = requestDepartureTime + requestTravelTime;

        var enrollmentHistory =
            await _unitOfWork.VehicleEnrollmentRepository.GetPageAsync(
                e =>
                    e.Vehicle.Id == entity.VehicleId &&
                    e.Id != entity.Id &&
                    e.DepartureTime >= DateTimeOffset.UtcNow.AddDays(-7),
                e => e.RouteAddressDetails,
                1, 200, cancellationToken);

        // Three cases are included:
        //
        // ---SD---------RD----------SA--->
        //                               time
        //
        // ---SD---------RA----------SA--->
        //                               time
        // 
        // ---RD-----SD-------SA-----RA--->
        //                               time
        // Where:
        //  RD - request enrollment departure time
        //  RA - request enrollment arrival time
        //  SD - datastore enrollment (S for store) departure time
        //  SA - datastore enrollment (S for store) arrival time

        var overlappingWithOtherEnrollments = enrollmentHistory.Items
            .Where(ve =>
            {
                var departureTime = ve.DepartureTime;

                var arrivalTime =
                    ve.DepartureTime +
                    ve.RouteAddressDetails
                        .Aggregate(
                            TimeSpan.Zero,
                            (sum, rad) => sum +
                                rad.TimeToNextAddress +
                                rad.CurrentAddressStopTime);

                return
                    (requestDepartureTime >= departureTime &&
                        requestDepartureTime <= arrivalTime) ||
                    (requestArrivalTime >= departureTime &&
                        requestArrivalTime <= arrivalTime) ||
                    (requestDepartureTime <= departureTime &&
                        requestArrivalTime >= arrivalTime);
            })
            .Any();

        if (overlappingWithOtherEnrollments)
        {
            throw new ValidationException(
                new List<ValidationFailure>
                {
                    new()
                    {
                        PropertyName = nameof(request.DepartureTime),
                        ErrorMessage = _localizer["Validation." +
                            "VehicleEnrollments.OverlapWithOther"]
                    }
                });
        }


        // Update entity and add to datastore.

        entity.DepartureTime = request.DepartureTime;
        entity.Currency = request.Currency;

        foreach (var rad in entity.RouteAddressDetails)
        {
            var correspondingRouteAddress = route.RouteAddresses
                .First(ra => ra.Id == rad.RouteAddressId);

            rad.TimeToNextAddress = request.RouteAddressDetails
                .First(rrad => rrad.RouteAddressGuid == rad.RouteAddress.Guid)
                .TimeToNextAddress;
            rad.CostToNextAddress = request.RouteAddressDetails
                .First(rrad => rrad.RouteAddressGuid == rad.RouteAddress.Guid)
                .CostToNextAddress;
            rad.CurrentAddressStopTime = request.RouteAddressDetails
                .First(rrad => rrad.RouteAddressGuid == rad.RouteAddress.Guid)
                .CurrentAddressStopTime;
        }

        entity = await _unitOfWork.VehicleEnrollmentRepository.UpdateOneAsync(
            entity, cancellationToken);

        await _unitOfWork.SaveAsync(cancellationToken);

        // Hydrate vehicle enrollment with address information

        var routeAddresses = await _unitOfWork.RouteAddressRepository
            .GetPageAsync(
                e =>
                    entity.RouteAddressDetails
                    .Select(rad => rad.RouteAddressId)
                    .Contains(e.Id),
                e => e.Address.City.Region.Country,
                1, entity.RouteAddressDetails.Count(),
                cancellationToken);

        foreach (var rad in entity.RouteAddressDetails)
        {
            rad.RouteAddress = routeAddresses.Items
                .First(ra => ra.Id == rad.RouteAddressId);
        }

        entity.RouteAddressDetails = entity.RouteAddressDetails
            .OrderBy(rad => rad.RouteAddress.Order)
            .ToArray();

        _unitOfWork.Dispose();

        return _mapper.Map<VehicleEnrollmentDto>(entity);
    }
}
