using MediatR;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Persistence;
using AutoMapper;
using cuqmbr.TravelGuide.Application.Common.Models;
using cuqmbr.TravelGuide.Application.Common.Extensions;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Services;
using cuqmbr.TravelGuide.Domain.Enums;

namespace cuqmbr.TravelGuide.Application.VehicleEnrollments
    .Queries.GetVehicleEnrollmentsPage;

public class GetVehicleEnrollmentsPageQueryHandler :
    IRequestHandler<GetVehicleEnrollmentsPageQuery, PaginatedList<VehicleEnrollmentDto>>
{
    private readonly UnitOfWork _unitOfWork;
    private readonly IMapper _mapper;

    private readonly SessionCurrencyService _sessionCurrencyService;
    private readonly CurrencyConverterService _currencyConverterService;

    public GetVehicleEnrollmentsPageQueryHandler(
        UnitOfWork unitOfWork,
        IMapper mapper,
        SessionCurrencyService sessionCurrencyService,
        CurrencyConverterService currencyConverterService)
    {
        _unitOfWork = unitOfWork;
        _mapper = mapper;
        _sessionCurrencyService = sessionCurrencyService;
        _currencyConverterService = currencyConverterService;
    }

    public async Task<PaginatedList<VehicleEnrollmentDto>> Handle(
        GetVehicleEnrollmentsPageQuery request,
        CancellationToken cancellationToken)
    {
        // TODO: Add search functionality or remove it
        var paginatedList = await _unitOfWork.VehicleEnrollmentRepository.GetPageAsync(
            e =>
                // (e.Name.ToLower().Contains(request.Search.ToLower()) ||
                // e.City.Region.Country.Name.ToLower().Contains(request.Search.ToLower())) &&
                (request.RouteGuid != null
                    ? e.Route.Guid == request.RouteGuid
                    : true) &&
                (request.VehicleGuid != null
                    ? e.Vehicle.Guid >= request.VehicleGuid
                    : true) &&
                (request.NumberOfAddressesGreaterThanOrEqual != null
                    ? 
                        e.RouteAddressDetails.Count() >=
                        request.NumberOfAddressesGreaterThanOrEqual
                    : true) &&
                (request.NumberOfAddressesLessThanOrEqual != null
                    ? 
                        e.RouteAddressDetails.Count() <=
                        request.NumberOfAddressesLessThanOrEqual
                    : true) &&
                (request.DepartureTimeGreaterThanOrEqual != null
                    ? e.DepartureTime >= request.DepartureTimeGreaterThanOrEqual
                    : true) &&
                (request.DepartureTimeLessThanOrEqual != null
                    ? e.DepartureTime <= request.DepartureTimeLessThanOrEqual
                    : true) &&
                (request.ArrivalTimeGreaterThanOrEqual != null
                    ? 
                        e.DepartureTime.AddSeconds(e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds +
                                rad.CurrentAddressStopTime.TotalSeconds)) >=
                        request.ArrivalTimeGreaterThanOrEqual
                    : true) &&
                (request.ArrivalTimeLessThanOrEqual != null
                    ? 
                        e.DepartureTime.AddSeconds(e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds +
                                rad.CurrentAddressStopTime.TotalSeconds)) <=
                        request.ArrivalTimeLessThanOrEqual
                    : true) &&
                (request.TravelTimeGreaterThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds) >=
                        request.TravelTimeGreaterThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.TravelTimeLessThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds) <=
                        request.TravelTimeLessThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.TimeMovingGreaterThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds) >=
                        request.TimeMovingGreaterThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.TimeMovingLessThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.TimeToNextAddress.TotalSeconds) <=
                        request.TimeMovingLessThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.TimeInStopsGreaterThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.CurrentAddressStopTime.TotalSeconds) >=
                        request.TimeInStopsGreaterThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.TimeInStopsLessThanOrEqual != null
                    ? 
                        e.RouteAddressDetails
                            .Sum(rad => rad.CurrentAddressStopTime.TotalSeconds) <=
                        request.TimeInStopsLessThanOrEqual.Value.TotalSeconds
                    : true) &&
                (request.CostGreaterThanOrEqual != null
                    ? e.RouteAddressDetails.Sum(rad => rad.CostToNextAddress) >=
                        request.CostGreaterThanOrEqual
                    : true) &&
                (request.CostLessThanOrEqual != null
                    ? e.RouteAddressDetails.Sum(rad => rad.CostToNextAddress) <=
                        request.CostLessThanOrEqual
                    : true) &&
                (request.Currency != null
                    ? e.Currency == request.Currency
                    : true),
            e => e.RouteAddressDetails,
            request.PageNumber, request.PageSize,
            cancellationToken);


        // Hydrate vehicle enrollment with address information

        var routeAddressIds = paginatedList.Items
            .SelectMany(ve => ve.RouteAddressDetails)
            .Select(rad => rad.RouteAddressId);

        var routeAddresses = await _unitOfWork.RouteAddressRepository
            .GetPageAsync(
                e =>
                    routeAddressIds.Contains(e.Id),
                e => e.Address.City.Region.Country,
                1, paginatedList.Items.Sum(e => e.RouteAddressDetails.Count()),
                cancellationToken);

        foreach (var vehicleEnrollment in paginatedList.Items)
        {
            foreach (var routeAddressDetail in
                    vehicleEnrollment.RouteAddressDetails)
            {
                routeAddressDetail.RouteAddress = routeAddresses.Items
                    .First(ra => ra.Id == routeAddressDetail.RouteAddressId);
            }
        }


        // Convert currency

        // TODO: Replace with AutoMapper Resolver

        if (!_sessionCurrencyService.Currency.Equals(Currency.Default))
        {
            foreach (var ve in paginatedList.Items)
            {
                foreach (var rad in ve.RouteAddressDetails)
                {
                    rad.CostToNextAddress = await _currencyConverterService
                        .ConvertAsync(rad.CostToNextAddress, ve.Currency,
                            _sessionCurrencyService.Currency, cancellationToken);
                }
                ve.Currency = _sessionCurrencyService.Currency;
            }
        }


        var mappedItems = _mapper
            .ProjectTo<VehicleEnrollmentDto>(paginatedList.Items.AsQueryable());

        mappedItems = QueryableExtension<VehicleEnrollmentDto>
            .ApplySort(mappedItems, request.Sort);

        _unitOfWork.Dispose();

        return new PaginatedList<VehicleEnrollmentDto>(
            mappedItems.ToList(),
            paginatedList.TotalCount, request.PageNumber,
            request.PageSize);
    }
}
