using cuqmbr.TravelGuide.Configuration.Persistence;
using cuqmbr.TravelGuide.Configuration.Application;
using cuqmbr.TravelGuide.Configuration.Infrastructure;
using cuqmbr.TravelGuide.Configuration.Identity;
using cuqmbr.TravelGuide.Configuration.Configuration;
using cuqmbr.TravelGuide.Configuration.Logging;
using cuqmbr.TravelGuide.Application.Common.Interfaces.Services;
using cuqmbr.TravelGuide.HttpApi.Services;
using cuqmbr.TravelGuide.HttpApi.Middlewares;
using cuqmbr.TravelGuide.HttpApi.Swashbuckle.OperationFilters;
using System.Net;
using Swashbuckle.AspNetCore.SwaggerUI;
using Microsoft.OpenApi.Models;
using System.Reflection;


var builder = WebApplication.CreateBuilder(args);

var services = builder.Services;


services.ConfigureConfiguration(args);

services.ConfigureLogging();

services.ConfigurePersistence();
services.ConfigureIdentity();
services.ConfigureInfrastructure();
services.ConfigureApplication();

services.AddScoped<SessionUserService, AspNetSessionUserService>();
services.AddScoped<SessionCultureService, AspNetSessionCultureService>();
services.AddScoped<SessionTimeZoneService, AspNetSessionTimeZoneService>();
services.AddScoped<SessionCurrencyService, AspNetSessionCurrencyService>();

services.AddControllers();


services.AddEndpointsApiExplorer();
services.AddSwaggerGen(options =>
{
    options.SwaggerDoc("v1", new OpenApiInfo()
    {
        Version = "v1",
        Title = "TravelGuide API",
        Description = "This document describes TravelGuide API"
    });

    options.EnableAnnotations();

    var xmlFilename = $"{Assembly.GetExecutingAssembly().GetName().Name}.xml";
    options.IncludeXmlComments(Path.Combine(AppContext.BaseDirectory, xmlFilename));

    options.OperationFilter<AuthorizationHeaderOperationFilter>();
    options.AddSecurityDefinition("Bearer", new OpenApiSecurityScheme
    {
        In = ParameterLocation.Header,
        Name = "Authorization",
        Description = "Bearer Authorization with Json Web Token " + 
            "(https://datatracker.ietf.org/doc/html/rfc7519)",
        Type = SecuritySchemeType.ApiKey
    });

    // Set Accept-Language header in Authorize window
    options.OperationFilter<AcceptLanguageHeaderOperationFilter>();
    options.AddSecurityDefinition("Accept-Language", new OpenApiSecurityScheme
    {
        In = ParameterLocation.Header,
        Name = "Accept-Language",
        Description = "Format: <ISO-639 language code>-<ISO-3166 " +
            "country/region code>. " + 
            "(https://datatracker.ietf.org/doc/html/rfc4646)",
        Type = SecuritySchemeType.ApiKey
    });

    // Set Accept-TimeZone header in Authorize window
    options.OperationFilter<AcceptTimeZoneHeaderOperationFilter>();
    options.AddSecurityDefinition("Accept-TimeZone", new OpenApiSecurityScheme
    {
        In = ParameterLocation.Header,
        Name = "Accept-TimeZone",
        Description = "Format: \\<Region\\>/\\<City\\>. Time Zone name " +
            "from IANA tz database (https://www.iana.org/time-zones).",
        Type = SecuritySchemeType.ApiKey
    });

    // Set Accept-Currency header in Authorize window
    options.OperationFilter<AcceptCurrencyHeaderOperationFilter>();
    options.AddSecurityDefinition("Accept-Currency", new OpenApiSecurityScheme
    {
        In = ParameterLocation.Header,
        Name = "Accept-Currency",
        Description = "ISO-4217 Currency Code.",
        Type = SecuritySchemeType.ApiKey
    });
});


services.AddScoped<ThreadCultureSetterMiddleware>();
services.AddScoped<GlobalExceptionHandlerMiddleware>();

services.AddHealthChecks();

builder.WebHost.ConfigureKestrel((context, options) =>
{
    // TODO: Make possible to configure using file/env/cli
    options.Configure()
        .Endpoint(new IPEndPoint(new IPAddress(0x00000000), 8080));
});



var app = builder.Build();

app.UseCors(builder => builder
    .AllowAnyOrigin()
    .AllowAnyMethod()
    .AllowAnyHeader());

app.MapControllers();

app.UseSwagger();
app.UseSwaggerUI(options =>
{
    options.DocExpansion(DocExpansion.None);
    options.DefaultModelRendering(ModelRendering.Model);
    options.DefaultModelExpandDepth(0);
    options.EnablePersistAuthorization();
});

app.UseMiddleware<ThreadCultureSetterMiddleware>();
app.UseMiddleware<GlobalExceptionHandlerMiddleware>();

app.MapHealthChecks("/health").WithOpenApi();



app.Run();
